# cli tests

import os
import shlex

import pytest
from click.testing import CliRunner

import rstms_cloudflare
from rstms_cloudflare import __version__, cli


def test_version():
    """Test reading version and module name"""
    assert rstms_cloudflare.__name__ == "rstms_cloudflare"
    assert __version__
    assert isinstance(__version__, str)


@pytest.fixture
def run():
    runner = CliRunner()

    env = os.environ.copy()
    env["TESTING"] = "1"

    def _run(cmd, **kwargs):
        assert_exit = kwargs.pop("assert_exit", 0)
        assert_exception = kwargs.pop("assert_exception", None)
        env.update(kwargs.pop("env", {}))
        kwargs["env"] = env
        result = runner.invoke(cli, cmd, **kwargs)
        if assert_exception is not None:
            assert isinstance(result.exception, assert_exception)
        elif result.exception is not None:
            raise result.exception from result.exception
        elif assert_exit is not None:
            assert result.exit_code == assert_exit, (
                f"Unexpected {result.exit_code=} (expected {assert_exit})\n"
                f"cmd: '{shlex.join(cmd)}'\n"
                f"output: {str(result.output)}"
            )
        return result

    return _run


def test_cli_no_args(run):
    result = run([])
    assert "Usage:" in result.output


def test_cli_help(run):
    result = run(["--help"])
    assert "Show this message and exit." in result.output


def test_cli_exception(run):
    cmd = ["--shell-completion", "and_now_for_something_completely_different"]

    with pytest.raises(RuntimeError) as exc:
        result = run(cmd)
    assert isinstance(exc.value, RuntimeError)

    # example of testing for expected exception
    result = run(cmd, assert_exception=RuntimeError)
    assert result.exception
    assert result.exc_info[0] == RuntimeError
    assert result.exception.args[0] == "cannot determine shell"

    with pytest.raises(AssertionError) as exc:
        result = run(cmd, assert_exception=ValueError)
    assert exc


def test_cli_exit(run):
    result = run(["--help"], assert_exit=None)
    assert result
    result = run(["--help"], assert_exit=0)
    assert result
    with pytest.raises(AssertionError):
        run(["--help"], assert_exit=-1)



"""Console script for rstms_cloudflare."""

import sys
from pathlib import Path

import click
import click.core

from .cloudflare import API
from .exception_handler import ExceptionHandler
from .shell import _shell_completion
from .version import __timestamp__, __version__

header = f"{__name__.split('.')[0]} v{__version__} {__timestamp__}"


def _ehandler(ctx, option, debug):
    ctx.obj = dict(ehandler=ExceptionHandler(debug))
    ctx.obj["debug"] = debug


@click.group(name="cloudflare")
@click.version_option(message=header)
@click.option("-d", "--debug", is_eager=True, is_flag=True, callback=_ehandler, help="debug mode")
@click.option(
    "--shell-completion",
    is_flag=False,
    flag_value="[auto]",
    callback=_shell_completion,
    help="configure shell completion",
)
@click.option("-q", "--quiet", is_flag=True)
@click.option("-r", "--raw", is_flag=True)
@click.option("-t", "--token", envvar="CLOUDFLARE_AUTH_TOKEN")
@click.option("-j", "--json", is_flag=True)
@click.option("-o", "--output", type=click.Path(writable=True, dir_okay=False, path_type=Path), help="output file")
@click.option("-i", "--id", "include_id", is_flag=True, help="include record ID in output")
@click.pass_context
def cli(ctx, debug, shell_completion, token, json, quiet, raw, include_id, output):
    if output:
        func = output.write_text
    else:
        func = click.echo
    ctx.obj = API(token=token, json=json, quiet=quiet, raw=raw, include_id=include_id, output_function=func)


@cli.command
@click.argument("domain", required=False)
@click.pass_obj
def dump(api, domain):
    """Output API data for one or all domains"""
    data = {}
    zones = api.get_zones()
    for zone in zones:
        if domain in [None, zone["name"]]:
            data[zone["name"]] = api.get_zone_records(zone["name"])
    if domain:
        if not data:
            raise ValueError(f"unknown domain: '{domain}'")

    api.json = True
    api.output(data)
    sys.exit(0 if data else -1)


def process_filter(arg, opt, name):
    if arg and opt:
        if arg == opt:
            opt = None
        else:
            raise RuntimeError(f"{name} option conflicts with argument")
    return arg or opt


@cli.command
@click.option("--delete", "delete_records", is_flag=True, help="delete selected records")
@click.option("--update-content", help="update selected record content")
@click.option("--update-ttl", help="update selected record TTL")
@click.option("-t", "--type", "filter_type", type=click.Choice(API.RECORD_TYPES + ["ID"]))
@click.option("-h", "--host", "filter_host", help="match host or /regex/")
@click.option("-c", "--content", "filter_content", help="match content or /regex/")
@click.option("-p", "--priority", type=int, help="filter by MX priority")
@click.argument("domain")
@click.argument("type", type=click.Choice(API.RECORD_TYPES + ["ID"]), required=False)
@click.argument("host", required=False)
@click.argument("content", required=False)
@click.pass_obj
def records(
    api,
    domain,
    type,
    host,
    content,
    filter_type,
    filter_host,
    filter_content,
    delete_records,
    priority,
    update_content,
    update_ttl,
):
    """select domain records"""

    if domain is None:
        domain = api.domain

    type = process_filter(type, filter_type, "type")
    host = process_filter(host, filter_host, "host")
    content = process_filter(content, filter_content, "content")

    records = api.select_records(domain, type, host, content, priority)

    if delete_records:
        records = api.delete_records(records)
    elif update_content or update_ttl:
        for record in records:
            if update_content:
                record["content"] = update_content
            if update_ttl:
                record["ttl"] = update_ttl
        records = api.update_records(records)
    else:
        records = api.format_records(records)

    api.output(records)
    if api.json:
        ret = 0
    else:
        ret = 0 if records else -1
    sys.exit(ret)


@cli.command
@click.pass_obj
def domains(api):
    """output registered domains"""
    zones = [zone["name"] for zone in api.get_zones()]
    if not api.json:
        zones = "\n".join(zones)
    api.output(zones)
    sys.exit(0 if zones else -1)


@cli.command
@click.argument("domain")
@click.pass_obj
def zone(api, domain):
    """output zone file"""

    zone_id = api.get_zone_id(domain)
    api.json = False
    api.output(api.client.zones.dns_records.export(zone_id))
    sys.exit(0)


@cli.command
@click.argument("domain")
@click.argument("type", type=click.Choice(API.RECORD_TYPES))
@click.argument("host")
@click.argument("content")
@click.option("-t", "--ttl", default=60)
@click.option("-p", "--priority", default=10)
@click.pass_obj
def add(api, domain, type, host, content, ttl, priority):
    """add a new record"""

    if type in ["A", "CNAME", "TXT"]:
        priority = None
    elif type == "MX":
        priority = priority or 0
    else:
        raise ValueError(f"Unsupported record type '{type}'")

    added = api.add_record(domain, type, host, content, ttl, priority)
    api.output(added)
    sys.exit(0)


@cli.command
@click.option("-p", "--priority")
@click.argument("domain")
@click.argument("type", type=click.Choice(API.RECORD_TYPES + ["ID"]))
@click.argument("host")
@click.argument("content", required=False)
@click.pass_obj
def delete(api, domain, type, host, content, priority):
    """delete matching record"""
    selected = api.select_records(domain, type, host, content, priority)
    deleted = api.delete_records(selected)
    api.output(deleted)
    sys.exit(0 if deleted else -1)


@cli.command
@click.option("-p", "--priority", type=int)
@click.option("-t", "--ttl", type=int)
@click.argument("domain")
@click.argument("type", type=click.Choice(API.RECORD_TYPES + ["ID"]))
@click.argument("host")
@click.argument("content", required=False)
@click.pass_obj
def update(api, domain, type, host, content, priority, ttl):
    """update matching record content"""
    selected = api.select_records(domain, type, host, content, priority)
    for record in selected:
        if content:
            record["content"] = content
        if priority and record["type"] == "MX":
            record["priority"] = content
        if ttl:
            record["ttl"] = ttl
    updated = api.update_records(selected)
    api.output(updated)
    sys.exit(0 if updated else -1)


if __name__ == "__main__":
    sys.exit(cli())



"""Cloudflare API Client"""

import json
import os
import re

import CloudFlare


class API:

    RECORD_TYPES = ["A", "AAAA", "CNAME", "MX", "NS", "SOA", "TXT", "SRV", "LOC"]
    MAX_ZONES = 128

    def __init__(self, token=None, json=False, quiet=False, raw=False, include_id=False, output_function=None):
        token = token or os.environ["CLOUDFLARE_AUTH_TOKEN"]
        self.client = CloudFlare.CloudFlare(token=token)
        self.json = json
        self.raw = raw
        self.quiet = quiet
        self.include_id = include_id
        self.output_function = output_function or print

    def output(self, data):
        if self.quiet:
            return
        if self.json:
            data = json.dumps(data, indent=2)
        elif not data:
            return
        self.output_function(data)

    def get_zones(self):
        return self.client.zones.get(params={"per_page": self.MAX_ZONES})

    def get_zone_id(self, domain):
        zones = self.get_zones()
        for z in zones:
            if z["name"] == domain:
                return z["id"]
        raise ValueError(f"unknown domain {domain}")

    def get_zone_records(self, domain):
        records = {}
        zone_id = self.get_zone_id(domain)
        records = self.client.zones.dns_records.get(zone_id, params={"per_page": self.MAX_ZONES})
        return records

    def parse_host(self, name, domain):
        if name in ["@", domain]:
            name = domain
        else:
            name = ".".join([name, domain])
        name = name.strip(".")
        return name

    def delete_records(self, records):
        deleted = [self.delete_record(record) for record in records]
        if self.json:
            return deleted
        else:
            return "\n".join(deleted)

    def delete_record(self, record):
        record_id = record["id"]
        zone_id = record["zone_id"]
        ret = self.client.zones.dns_records.delete(zone_id, record_id)
        if self.json:
            return ret
        else:
            return ret["id"]

    def add_record(self, domain, type, host, content, ttl=None, priority=None):
        zone_id = self.get_zone_id(domain)
        host = self.parse_host(host, domain)
        record = dict(type=type, name=host, content=content, ttl=ttl)
        if type == "MX":
            record["priority"] = priority or 0
        ret = self.client.zones.dns_records.post(zone_id, data=record)
        if self.json:
            return ret
        else:
            return ret["id"]

    def update_records(self, records):
        updated = [self.update_record(record) for record in records]
        if self.json:
            return updated
        else:
            return "\n".join(updated)

    def update_record(self, record):
        dns_record_id = record["id"]
        zone_id = record["zone_id"]
        update = dict(
            name=record["name"],
            type=record["type"],
            content=record["content"],
            ttl=record["ttl"],
            proxied=record["proxied"],
        )
        ret = self.client.zones.dns_records.patch(zone_id, dns_record_id, data=update)

        if self.json:
            return ret
        else:
            return ret["id"]

    def format_host(self, record):
        domain = record["zone_name"]
        host = record["name"]
        if host.endswith(domain):
            host = host[: -1 - len(domain)]
        if not host:
            host = "@"
        return host

    def format_record(self, record):
        if self.json:
            if self.raw:
                out = record
            else:
                out = dict(
                    domain=record["zone_name"],
                    name=self.format_host(record),
                    content=record["content"],
                    type=record["type"],
                    ttl=record["ttl"],
                    priority=record.get("priority", None),
                )
                if record["type"] == "MX":
                    out["priority"] = record["priority"]
                if self.include_id:
                    out["id"] = record["id"]
        else:
            out = ""
            if self.include_id:
                out += f"{record['id']} "
            out += f"{record['type']} "
            if record["type"] == "MX":
                out += f"{record['priority']} "
            if record["type"] == "TXT":
                content = '"' + record["content"] + '"'
            else:
                content = record["content"]
            out += f"{record['name']} {content} {record['ttl']}"
        return out

    def format_records(self, records):
        formatted = [self.format_record(r) for r in records]
        if self.json:
            if self.include_id and not self.raw:
                formatted = {record["id"]: record for record in formatted}
                [record.pop("id") for record in formatted.values()]
            return formatted
        else:
            return "\n".join(formatted)

    def is_selected(self, pattern, text):
        if pattern is None:
            return True
        elif pattern.startswith("/"):
            return bool(re.match(pattern.strip("/"), text))
        else:
            return pattern == text

    def select_records(self, domain, type=None, host=None, content=None, priority=None):

        if host and type != "ID" and not host.startswith("/"):
            host = self.parse_host(host, domain)

        records = self.get_zone_records(domain)

        selected = []

        if type == "ID":
            for record in records:
                if host == record["id"]:
                    return [record]
            return []

        for record in records:
            if not self.is_selected(type, record["type"]):
                continue
            if not self.is_selected(host, record["name"]):
                continue
            if not self.is_selected(content, record["content"]):
                continue
            if record["type"] == "MX":
                if priority is not None:
                    if int(priority) != int(record["priority"]):
                        continue
            selected.append(record)

        return selected
